EXPORTS
    Mat_GetLibraryVersion
    Mat_SetVerbose
    Mat_SetDebug
    Mat_Critical
    Mat_Error
    Mat_Help
    Mat_LogInit
    Mat_LogClose
    Mat_LogInitFunc
    Mat_Message
    Mat_DebugMessage
    Mat_VerbMessage
    Mat_Warning
    Mat_SizeOf
    Mat_SizeOfClass
    Mat_CreateVer
    Mat_Close
    Mat_Open
    Mat_GetFilename
    Mat_GetVersion
    Mat_Rewind
    Mat_VarCalloc
    Mat_VarCreate
    Mat_VarCreateStruct
    Mat_VarDelete
    Mat_VarDuplicate
    Mat_VarFree
    Mat_VarGetCell
    Mat_VarGetCells
    Mat_VarGetCellsLinear
    Mat_VarGetSize
    Mat_VarGetNumberOfFields
    Mat_VarAddStructField
    Mat_VarGetStructFieldByIndex
    Mat_VarGetStructFieldByName
    Mat_VarGetStructField
    Mat_VarGetStructFieldnames
    Mat_VarGetStructs
    Mat_VarGetStructsLinear
    Mat_VarPrint
    Mat_VarRead
    Mat_VarReadData
    Mat_VarReadDataAll
    Mat_VarReadDataLinear
    Mat_VarReadInfo
    Mat_VarReadNext
    Mat_VarReadNextInfo
    Mat_VarSetCell
    Mat_VarSetStructFieldByIndex
    Mat_VarSetStructFieldByName
    Mat_VarWrite
    Mat_VarWriteInfo
    Mat_VarWriteData
    Mat_CalcSingleSubscript
    Mat_CalcSubscripts
